from taggingbackends.data.labels import Labels
from taggingbackends.data.dataset import LarvaDataset
from taggingbackends.explorer import check_permissions
from maggotuba.models.trainers import MaggotTrainer, MultiscaleMaggotTrainer, new_generator
import json
import glob

def train_model(backend, layers=1, pretrained_model_instance="default", subsets=(1, 0, 0), **kwargs):
    # make_dataset generated or moved the larva_dataset file into data/interim/{instance}/
    #larva_dataset_file = backend.list_interim_files("larva_dataset_*.hdf5") # recursive
    larva_dataset_file = glob.glob(str(backend.interim_data_dir() / "larva_dataset_*.hdf5")) # not recursive (faster)
    assert len(larva_dataset_file) == 1
    # subsets=(1, 0, 0) => all data are training data; no validation or test subsets
    dataset = LarvaDataset(larva_dataset_file[0], new_generator(), subsets=subsets, **kwargs)
    labels = dataset.labels
    assert 0 < len(labels)
    labels = labels if isinstance(labels[0], str) else [s.decode() for s in labels]
    # copy and load the pretrained model into the model instance directory
    if isinstance(pretrained_model_instance, str):
        config_file = import_pretrained_model(backend, pretrained_model_instance)
        model = MaggotTrainer(config_file, labels, layers)
    else:
        pretrained_model_instances = pretrained_model_instance
        config_files = import_pretrained_models(backend, pretrained_model_instances)
        model = MultiscaleMaggotTrainer(config_files, labels, layers)
    # fine-tune and save the model
    model.train(dataset)
    print(f"saving model \"{backend.model_instance}\"")
    model.save()

# TODO: merge the below two functions

def import_pretrained_model(backend, pretrained_model_instance):
    pretrained_autoencoder_dir = backend.project_dir / "pretrained_models" / pretrained_model_instance
    config_file = None
    for file in pretrained_autoencoder_dir.iterdir():
        if not file.is_file():
            continue
        dst = backend.model_dir() / file.name
        if file.name.endswith("config.json"):
            with open(file) as f:
                config = json.load(f)
            dir = backend.model_dir().relative_to(backend.project_dir)
            config["log_dir"] = str(dir)
            # optional updates?
            #config["project_dir"] = config["exp_folder"] = str(dir)
            #config["exp_name"] = backend.model_instance
            #config["config"] = str(dir / os.path.basename(config["config"]))
            with open(dst, "w") as f:
                json.dump(config, f, indent=2)
            assert config_file is None
            config_file = dst
        else:
            with open(file, "rb") as i:
                with open(dst, "wb") as o:
                    o.write(i.read())
        check_permissions(dst)
    return config_file

def import_pretrained_models(backend, model_instances):
    config_files = []
    for pretrained_model_instance in model_instances:
        pretrained_autoencoder_dir = backend.project_dir / "pretrained_models" / pretrained_model_instance
        encoder_dir = backend.model_dir() / pretrained_model_instance
        encoder_dir.mkdir(exist_ok=True)
        config_file = None
        for file in pretrained_autoencoder_dir.iterdir():
            if not file.is_file():
                continue
            dst = encoder_dir / file.name
            if file.name.endswith("config.json"):
                with open(file) as f:
                    config = json.load(f)
                dir = encoder_dir.relative_to(backend.project_dir)
                config["log_dir"] = str(dir)
                with open(dst, "w") as f:
                    json.dump(config, f, indent=2)
                assert config_file is None
                config_file = dst
            else:
                with open(file, "rb") as i:
                    with open(dst, "wb") as o:
                        o.write(i.read())
            check_permissions(dst)
        assert config_file is not None
        config_files.append(config_file)
    return config_files


from taggingbackends.main import main

if __name__ == "__main__":
    main(train_model)
